# # pylint: disable=missing-function-docstring
# # pylint: disable=missing-class-docstring
# # pylint: disable=line-too-long
# '''
#     Methods for managing drawio nodes that are wrapped in an object tag.

#     ----------

#     Meta
#     ----------
#     `author`: Colemen Atwood
#     `created`: 06-04-2022 15:45:00
#     `memberOf`: drawio
#     `name`: onode
# '''

# from lxml import etree as _etree
# import utils.string_utils as _csu
# import utils.dict_utils as _obj
# from utils.drawio.nodeBase import NodeBase as _NodeBase
# import utils.drawio.diagram_utils as _dia

# def new_onode(tree,diagram,id:str=None,**kwargs):
#     '''
#         Creates a new element for an object node.

#         ----------

#         Keyword Arguments
#         -------------------------
#         [`x`=0] {int|str}
#             The initial x coordinate of the node
#         [`y`=0] {int|str}
#             The initial y coordinate of the node
#         [`w`=120] {int|str}
#             The initial width of the node
#         [`h`=60] {int|str}
#             The initial height of the node

#         Arguments
#         -------------------------
#         `tree` {object}
#             A reference to the lxml tree object.
#         `diagram` {object}
#             A reference to the Diagram instance this node is added to.
#         [`id`=None] {string}
#             The optional id of the node, if not provided, a random one is generated.

#         Return {Onode}
#         ----------------------
#         An Onode instance.

#         Meta
#         ----------
#         `author`: Colemen Atwood
#         `created`: 05-27-2022 11:13:10
#         `memberOf`: onode
#         `version`: 1.0
#         `method_name`: new_onode
#     '''

#     x = _obj.get_kwarg(['x'],"0",(int,str),**kwargs)
#     y = _obj.get_kwarg(['y'],"0",(int,str),**kwargs)
#     width = _obj.get_kwarg(['width','w'],"120",(int,str),**kwargs)
#     height = _obj.get_kwarg(['height','h'],"60",(int,str),**kwargs)

#     o = _etree.SubElement(diagram.dia_root, 'object')
#     id = _csu.gen.rand() if id is None else id
#     o.attrib['id'] = id

#     mxCell = _etree.SubElement(o, 'mxCell')
#     # mxCell.attrib['style']="shape=hexagon;perimeter=hexagonPerimeter2;whiteSpace=wrap;html=1;fixedSize=1;size=10;fillColor=#76608a;strokeColor=#432D57;fontColor=#ffffff;"
#     mxCell.attrib['style']="rounded=0;whiteSpace=wrap;html=1;"
#     mxCell.attrib['vertex']="1"
#     mxCell.attrib['parent']="1"

#     mxGeo = _etree.SubElement(mxCell, 'mxGeometry')
#     mxGeo.attrib['x']=str(x)
#     mxGeo.attrib['y']=str(y)
#     mxGeo.attrib['width']=str(width)
#     mxGeo.attrib['height']=str(height)
#     mxGeo.attrib['as']="geometry"

#     return Onode(tree,o,diagram)

# class Onode(_NodeBase):
#     def __init__(self,tree,element=None,diagram=None):
#         super().__init__(tree,element,diagram)
#         self.settings = {}
#         self.tree = tree
#         self.element = element
#         self.data = {
#             "node_type":"onode",
#             "lxml":"",
#             "xml":"",
#             "coords":{
#                 "x":None,
#                 "y":None,
#                 "tlc":[],
#                 "trc":[],
#                 "brc":[],
#                 "blc":[],
#             },
#             "attributes":{},
#             "mxcell":{
#                 "attributes":{},
#                 "style":{},
#                 "vertex":None,
#                 "parent":None,
#             },
#             "mxgeometry":{
#                 "x":None,
#                 "y":None,
#                 "width":None,
#                 "height":None,
#                 "as":None,
#             },
#         }
#         self._from_element()

#     def _from_element(self):
#         element = self.element
#         if element is not None:
#             self.data['attributes'] = _dia.attrib_to_dict(element.attrib)
#             if 'tags' in self.data['attributes']:
#                 self.data['attributes']['tags'] = self.data['attributes']['tags'].split(",")

#             # @Mstep [] parse the mxcell attributes
#             mxcell = element.xpath('mxCell')
#             if len(mxcell) > 0:
#                 self.data['mxcell']['attributes'] = _dia.attrib_to_dict(mxcell[0].attrib)
#                 # style = _dia.style_to_dict(self.data['mxcell']['attributes']['style'])
#                 # print(f"style:{style}")

#                 # @Mstep [] retrieve and parse the mxGeometry attributes.
#                 mxgeometry = mxcell[0].xpath('mxGeometry')
#                 self.data['mxgeometry'] = _dia.attrib_to_dict(mxgeometry[0].attrib)

#                 self.data['mxcell']['attributes']['style'] = _dia.style_to_dict(self.data['mxcell']['attributes']['style'])


#             return self.data

#     def has_tag(self,value):
#         '''
#             Check if this object node contains a matching tag in the tags attribute.

#             ----------

#             Arguments
#             -------------------------
#             `value` {string|list}
#                 The tag to search for.

#             Return {bool}
#             ----------------------
#             True upon success, false otherwise.

#             Meta
#             ----------
#             `author`: Colemen Atwood
#             `created`: 05-27-2022 10:56:33
#             `memberOf`: onode
#             `version`: 1.0
#             `method_name`: has_tag
#             # @xxx [05-27-2022 10:59:21]: documentation for has_tag
#         '''

#         if isinstance(value,(str)):
#             value = [value]
#         for v in value:
#             if 'tags' in self.data['attributes']:
#                 if v in self.data['attributes']['tags']:
#                     return True
#             return False

#     def set_tag(self,tag=None):
#         '''
#             Adds a new tag to the tags attribute of the object node.

#             ----------

#             Arguments
#             -------------------------
#             `tag` {string|None}
#                 The new tag to add.
#                 if None, all tags will be removed.

#             Meta
#             ----------
#             `author`: Colemen Atwood
#             `created`: 05-27-2022 10:59:30
#             `memberOf`: onode
#             `version`: 1.0
#             `method_name`: set_tag
#         '''

#         if tag is None:
#             self.data['attributes']['tags'] = []
#             return
            
#         if 'tags' in self.data['attributes']:
#             if isinstance(self.data['attributes']['tags'],(str)):
#                 self.data['attributes']['tags'] = self.data['attributes']['tags'].split(",")

#         if 'tags' not in self.data['attributes']:
#             self.data['attributes']['tags'] = []

#         self.data['attributes']['tags'].append(tag)
#         self.element.attrib['tags'] = ','.join(self.data['attributes']['tags'])

#     def remove_tag(self,tag):
#         '''
#             Remove a tag from the object node's tags attribute

#             ----------

#             Arguments
#             -------------------------
#             `tag` {string}
#                 The tag to remove.

#             Return {bool}
#             ----------------------
#             return_description

#             Meta
#             ----------
#             `author`: Colemen Atwood
#             `created`: 05-27-2022 11:01:49
#             `memberOf`: onode
#             `version`: 1.0
#             `method_name`: remove_tag
#         '''

#         new_tags = []
#         if 'tags' in self.data['attributes']:
#             for t in self.data['attributes']['tags']:
#                 if t != tag:
#                     new_tags.append(t)
#         self.data['attributes']['tags'] = new_tags
#         self.element.attrib['tags'] = ','.join(new_tags)



# # pylint: disable=missing-function-docstring
# # pylint: disable=missing-class-docstring
# # pylint: disable=line-too-long
# '''
#     Methods for managing connector nodes.

#     ----------

#     Meta
#     ----------
#     `author`: Colemen Atwood
#     `created`: 06-04-2022 16:21:17
#     `memberOf`: drawio
#     `name`: connector
# '''


# from lxml import etree as _etree
# import utils.string_utils as _csu
# from utils.drawio.nodeBase import NodeBase as _NodeBase
# from utils.drawio.diagram import Diagram as _Diagram
# import utils.drawio.diagram_utils as _dia




# def new_connector(tree:_etree,diagram:_Diagram,source:str,target:str):
#     '''
#         Create a new connector node in the diagram.

#         ----------

#         Arguments
#         -------------------------
#         `tree` {etree}
#             The node tree of the diagram.

#         `diagram` {Diagram}
#             A reference to the Diagram instance that this node belonds to.

#         `source` {str}
#             The id of the node that the connector starts from.

#         `target` {str}
#             The id of the node that the connector points to.

#         Return {Connector}
#         ----------------------
#         An instance of the Connector class.

#         Meta
#         ----------
#         `author`: Colemen Atwood
#         `created`: 06-04-2022 16:40:21
#         `memberOf`: connector
#         `version`: 1.0
#         `method_name`: new_connector
#         * @xxx [06-04-2022 16:42:20]: documentation for new_connector
#     '''


#     o = _etree.SubElement(diagram.dia_root, 'mxCell')
#     # id = _csu.gen.rand() if id is None else id
#     o.attrib['id'] = _csu.gen.rand()
#     o.attrib['style']="edgeStyle=entityRelationEdgeStyle;rounded=0;orthogonalLoop=1;jettySize=auto;html=1;exitX=1;exitY=0.5;exitDx=0;exitDy=0;entryX=0;entryY=0.5;entryDx=0;entryDy=0;"
#     o.attrib['edge']="1"
#     o.attrib['parent']="1"
#     o.attrib['source']=source
#     o.attrib['target']=target
    
#     mxgeo = _etree.SubElement(o, 'mxGeometry')
#     mxgeo.attrib['relative'] = "1"
#     mxgeo.attrib['as'] = "geometry"
    
#     return Connector(tree,o,diagram)

# class Connector(_NodeBase):
#     '''
#         This class is used for managing connector arrows in the drawio diagrams.

#         ----------

#         Arguments
#         -------------------------
#         `tree` {etree}
#             The node tree of the diagram.

#         `diagram` {Diagram}
#             A reference to the Diagram instance that this node belonds to.

#         `source` {str}
#             The id of the node that the connector starts from.

#         `target` {str}
#             The id of the node that the connector points to.

#         Meta
#         ----------
#         `author`: Colemen Atwood
#         `created`: 06-04-2022 16:36:04
#         `memberOf`: connector
#         `version`: 1.0
#         `name`: Connector
#         * @xxx [06-04-2022 16:40:10]: documentation for Connector
#     '''


#     def __init__(self,tree:_etree,element=None,diagram:_Diagram=None):
#         super().__init__(tree,element,diagram)
#         self.settings = {}
#         self.data = {
#             "node_type":"connector",
#             "attributes":{},
#         }
#         self._from_element()

#     def _from_element(self):
#         element = self.element
#         if element is not None:
#             self.data['attributes'] = _dia.attrib_to_dict(element.attrib)
#             return self.data



'''
    A module of utility methods used for formatting strings.

    ----------

    Meta
    ----------
    `author`: Colemen Atwood
    `created`: 12-09-2021 09:00:27
    `memberOf`: string_utils
'''


# import json
# import hashlib
import logging as _logging
# from typing import Union as _Union
# import re as _re
# import os as _os
# import urllib.parse

import utils.dict_utils as _obj
from pyparsing import QuotedString
# from pyparsing import And, Suppress,Dict, Word, Literal, Group, Optional, ZeroOrMore, OneOrMore, Regex, restOfLine, alphanums,nums, printables, string, CaselessKeyword,nestedExpr,ParseException,quotedString,removeQuotes,originalTextFor,delimitedList,QuotedString

# from utils.dict_utils.dict_utils import get_kwarg as _obj.get_kwarg

logger = _logging.getLogger(__name__)

_numeric_char_codes = {
    "symbols":{
        ";":"&#59;",
        "#":"&#35;",
        "&":"&#38;",
        "!":"&#33;",
        "\"":"&#34;",
        "$":"&#36;",
        "%":"&#37;",
        "'":"&#39;",
        "(":"&#40;",
        ")":"&#41;",
        "*":"&#42;",
        "+":"&#43;",
        ",":"&#44;",
        "-":"&#45;",
        ".":"&#46;",
        "/":"&#47;",
        ":":"&#58;",
        "<":"&#60;",
        "=":"&#61;",
        ">":"&#62;",
        "?":"&#63;",
        "@":"&#64;",
        "[":"&#91;",
        "\\":"&#92;",
        "]":"&#93;",
        "^":"&#94;",
        "_":"&#95;",
        "`":"&#96;",
        "{":"&#123;",
        "|":"&#124;",
        "}":"&#125;",
        "~":"&#126;"
    },
    "numeric":{
        "0":"&#48;",
        "1":"&#49;",
        "2":"&#50;",
        "3":"&#51;",
        "4":"&#52;",
        "5":"&#53;",
        "6":"&#54;",
        "7":"&#55;",
        "8":"&#56;",
        "9":"&#57;"
    },
    "upper_case":{
        "A":"&#65;",
        "B":"&#66;",
        "C":"&#67;",
        "D":"&#68;",
        "E":"&#69;",
        "F":"&#70;",
        "G":"&#71;",
        "H":"&#72;",
        "I":"&#73;",
        "J":"&#74;",
        "K":"&#75;",
        "L":"&#76;",
        "M":"&#77;",
        "N":"&#78;",
        "O":"&#79;",
        "P":"&#80;",
        "Q":"&#81;",
        "R":"&#82;",
        "S":"&#83;",
        "T":"&#84;",
        "U":"&#85;",
        "V":"&#86;",
        "W":"&#87;",
        "X":"&#88;",
        "Y":"&#89;",
        "Z":"&#90;"
    },
    "lower_case":{

        "a":"&#97;",
        "b":"&#98;",
        "c":"&#99;",
        "d":"&#100;",
        "e":"&#101;",
        "f":"&#102;",
        "g":"&#103;",
        "h":"&#104;",
        "i":"&#105;",
        "j":"&#106;",
        "k":"&#107;",
        "l":"&#108;",
        "m":"&#109;",
        "n":"&#110;",
        "o":"&#111;",
        "p":"&#112;",
        "q":"&#113;",
        "r":"&#114;",
        "s":"&#115;",
        "t":"&#116;",
        "u":"&#117;",
        "v":"&#118;",
        "w":"&#119;",
        "x":"&#120;",
        "y":"&#121;",
        "z":"&#122;"
    }
}



def regex(value:str)->str:
    '''
        Escapes regex special characters.

        ----------

        Arguments
        -------------------------
        `value` {str}
            The string to escape.

        Return {str}
        ----------------------
        The formatted string.

        Meta
        ----------
        `author`: Colemen Atwood
        `created`: 12-09-2021 08:46:32
        `memberOf`: string_format
        `version`: 1.0
        `method_name`: escape_regex
    '''
    regex_chars = ["\\", "^", "$", "{", "}", "[", "]", "(", ")", ".", "*", "+", "?", "<", ">", "-", "&"]
    for char in regex_chars:
        value = value.replace(char, f"\\{char}")
    return value
escape_regex = regex

def encode_quotes(value:str)->str:
    '''
        Encodes single and double quotes within the value to &apos; or &quot; respectively.

        ----------

        Arguments
        -------------------------
        `value` {str}
            The string to encode

        Return {str}
        ----------------------
        The encoded string.

        Meta
        ----------
        `author`: Colemen Atwood
        `created`: 12-09-2021 08:15:52
        `memberOf`: string_format
        `version`: 1.0
        `method_name`: encode_quotes
    '''
    value = value.replace("'", "&apos;")
    value = value.replace('"', "&quot;")
    return value

def decode_quotes(value:str)->str:
    '''
        Decodes single and double quotes within the value from &apos; or &quot; respectively.

        ----------

        Arguments
        -------------------------
        `value` {str}
            The string to decode

        Return {str}
        ----------------------
        The decoded string.

        Meta
        ----------
        `author`: Colemen Atwood
        `created`: 12-09-2021 08:15:52
        `memberOf`: string_format
        `version`: 1.0
        `method_name`: encode_quotes
    '''
    value = value.replace("&apos;", "'")
    value = value.replace("&quot;", '"')
    return value

def reverse_sanitize_quotes(string):
    orig_list = False
    if isinstance(string, (list)):
        orig_list = True
    if isinstance(string, (str)):
        string = [string]

    new_list = []
    for item in string:
        item = item.replace("&apos_", "'")
        item = item.replace("&quot_", '"')
        new_list.append(item)

    if len(new_list) == 1 and orig_list is False:
        return new_list[0]

    return new_list

def sanitize_quotes(string):
    orig_list = False
    if isinstance(string, (list)):
        orig_list = True
    if isinstance(string, (str)):
        string = [string]

    new_list = []
    for item in string:
        if isinstance(item, (str)):
            item = item.replace("'", "&apos_")
            item = item.replace('"', "&quot_")
        new_list.append(item)

    if len(new_list) == 1 and orig_list is False:
        return new_list[0]

    return new_list

def _get_numeric_charcode(char):
    for _,v in _numeric_char_codes.items():
        for c,code in v.items():
            if c == char:
                return code
    return None

def _get_numeric_charcode_chars(cat:str)->list:
    output = []
    if cat in _numeric_char_codes:
        for k,_ in _numeric_char_codes[cat].items():
            output.append(k)
    return output

def escape_charcode(value:str,reverse=False,**kwargs)->str:
    '''
        Escape character in a string with their character codes.

        ----------

        Arguments
        -------------------------
        `value` {str}
            The string to escape.

        [`reverse`=False] {bool}
            If True, it will unescape the charcoded characters.

        Keyword Arguments
        -------------------------
        `chars` {list|str}
            A list or string of characters to escape.

        [`symbols`=True] {bool}
            If True, all symbols will be escaped.

        [`numeric`=False] {bool}
            If True, all numeric will be escaped.

        [`upper_case`=False] {bool}
            If True, all upper_case will be escaped.

        [`lower_case`=False] {bool}
            If True, all lower_case will be escaped.

        Return {str}
        ----------------------
        The escaped string.

        Meta
        ----------
        `author`: Colemen Atwood
        `created`: 06-04-2022 09:36:42
        `memberOf`: string_format
        `version`: 1.0
        `method_name`: escape_charcode
        * @xxx [06-04-2022 09:38:48]: documentation for escape_charcode
    '''

    if reverse is True:
        return charcode_unescape(value)

    chars = _obj.get_kwarg(["chars"], None, (list,str), **kwargs)
    symbols = _obj.get_kwarg(["symbols"], True, (bool), **kwargs)
    numeric = _obj.get_kwarg(["numeric"], False, (bool), **kwargs)
    upper_case = _obj.get_kwarg(["upper_case"], False, (bool), **kwargs)
    lower_case = _obj.get_kwarg(["lower_case"], False, (bool), **kwargs)

    if chars is not None:
        if isinstance(chars,(str)):
            chars = chars.split()
        for c in chars:
            code = _get_numeric_charcode(c)
            value = value.replace(c,code)
        return value

    chars = []

    if symbols is True:
        chars = chars + _get_numeric_charcode_chars('symbols')
    if numeric is True:
        chars = chars + _get_numeric_charcode_chars('numeric')
    if upper_case is True:
        chars = chars + _get_numeric_charcode_chars('upper_case')
    if lower_case is True:
        chars = chars + _get_numeric_charcode_chars('lower_case')

    for c in chars:
        code = _get_numeric_charcode(c)
        value = value.replace(c,code)

    return value
charcode = escape_charcode

def charcode_unescape(value:str)->str:
    '''
        Unescape the character codes from escape_charcode.

        ----------

        Arguments
        -------------------------
        `value` {str}
            The string to unescape.

        Return {str}
        ----------------------
        The unescaped string.

        Meta
        ----------
        `author`: Colemen Atwood
        `created`: 06-04-2022 09:39:02
        `memberOf`: string_format
        `version`: 1.0
        `method_name`: charcode_unescape
        * @xxx [06-04-2022 09:39:50]: documentation for charcode_unescape
    '''

    for _,v in _numeric_char_codes.items():
        for c,code in v.items():
            value = value.replace(code,c)
    return value

def quoted_commas(value,escape_value="__ESCAPED_COMMA__",reverse=False):
    '''
        Escape commas that are located within quotes.

        ----------

        Arguments
        -------------------------
        `value` {str}
            The string to search within
        [`escape_value`='__ESCAPED_COMMA__] {str}
            The value to replace commas with.
        `reverse` {bool}
            if True it will replace the escaped commas with actual commas.

        Return {str}
        ----------------------
        The string with escaped commas.

        Meta
        ----------
        `author`: Colemen Atwood
        `created`: 06-01-2022 06:54:48
        `memberOf`: parse_sql
        `version`: 1.0
        `method_name`: escape_quoted_commas
        @xxx [06-01-2022 06:57:45]: documentation for escape_quoted_commas
    '''

    if reverse is True:
        return value.replace(escape_value,",")

    sql_qs = QuotedString("'", esc_quote="''")
    quote = sql_qs.search_string(value)
    if len(quote) > 0:
        quote = quote.asList()
        for q in quote:
            if len(q) == 1:
                q = q[0]
            esc = q.replace(",",escape_value)
            value = value.replace(q,esc)

    return value
escape_quoted_commas = quoted_commas

def quoted_chars(value,reverse=False):
    '''
        Escape characters that can effect parsing which are located within quotes.

        ----------

        Arguments
        -------------------------
        `value` {str}
            The string to search within

        `reverse` {bool}
            if True it will reverse the escaped chars with their actual chars.

        Return {str}
        ----------------------
        The string with escaped chars.

        Meta
        ----------
        `author`: Colemen Atwood
        `created`: 06-01-2022 06:54:48
        `memberOf`: parse_sql
        `version`: 1.0
        `method_name`: escape_quoted_chars
        @xxx [06-01-2022 06:57:45]: documentation for escape_quoted_chars
    '''

    escapes = [
        [",","__&#44__"],
        [";","__&#59__"],
        ["(","__&#40__"],
        [")","__&#41__"],
        ["`","__&#96__"],
        ['"',"__&#34__"],
        ["'","__&#39__"],
    ]


    if reverse is True:
        for e in escapes:
            value = value.replace(e[1],e[0])
        return value

    for e in escapes:
        sql_qs = QuotedString("'", esc_quote="''")
        quote = sql_qs.search_string(value)
        if len(quote) > 0:
            quote = quote.asList()
            # print(f"quote: {quote}")
            for q in quote:
                if len(q) == 1:
                    q = q[0]
                esc = q.replace(e[0],e[1])
                value = value.replace(q,esc)
    # print(sql_qs.search_string(value))
    return value
escape_quoted_chars = quoted_chars














# pylint: disable=missing-function-docstring
# pylint: disable=missing-class-docstring
# pylint: disable=line-too-long
# pylint: disable=unused-import

'''
    A composite library of utility methods used for all sorts of shit.

    ----------

    Meta
    ----------
    `author`: Colemen Atwood
    `created`: 06-03-2022 10:28:36
    `memberOf`: sql
    `version`: 1.0
'''


import utils.list_utils as arr
import utils.dict_utils as obj
import utils.random_utils as rand
import utils.string_utils as string
import utils.type_utils as types
import utils.file_utils as file
import utils.directory_utils as dirs
import utils.database_utils as db
# import utils.drawio as drawio








# # pylint: disable=missing-function-docstring
# # pylint: disable=missing-class-docstring
# # pylint: disable=line-too-long
# '''
#     Methods for managing drawio drawings.

#     ----------

#     Meta
#     ----------
#     `author`: Colemen Atwood
#     `created`: 06-04-2022 15:45:00
#     `memberOf`: drawio
#     `name`: drawing
# '''


# from lxml import etree
# # import json as _json

# import utils.file_utils as _f
# import utils.dict_utils as _obj
# # import utils.drawio.diagram_utils as _dia

# # import objectUtils as obj
# from utils.drawio.nodeBase import NodeBase as _NodeBase
# # from nodeBase import NodeBase
# # from utils.drawio import Connector as _connector
# # from connector import Connector
# # from onode import Onode
# from typing import TypeVar as _TypeVar
# from utils.drawio.diagram import new_diagram as _new_diagram
# from utils.drawio.diagram import Diagram as _Diagram
# # from diagram import new_diagram,Diagram

# _diagram_type = _TypeVar('_diagram_type', bound=_Diagram)

# def new_drawing(**kwargs):
#     '''
#         Creates a new draw.io drawing.

#         ----------

#         Keyword Arguments
#         -------------------------
#         [`file_path`=None] {string}
#             The path to where the drawing should be saved.

#         Return {Drawing}
#         ----------------------
#         An instance of the Drawing class

#         Meta
#         ----------
#         `author`: Colemen Atwood
#         `created`: 05-27-2022 12:22:38
#         `memberOf`: drawing
#         `version`: 1.0
#         `method_name`: new_drawing
#     '''

#     file_path = _obj.get_kwarg(['file_path','path'],None,(str),**kwargs)

#     mxfile = etree.Element("mxfile")
#     mxfile.attrib['host'] = "Electron"
#     mxfile.attrib['modified'] = "2022-05-26T18:37:31.077Z"
#     mxfile.attrib['agent'] = "5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) draw.io/14.9.6 Chrome/89.0.4389.128 Electron/12.0.16 Safari/537.36"
#     mxfile.attrib['etag'] = "AdGKFnLcEHsRAx8fGjjM"
#     mxfile.attrib['compressed'] = "false"
#     mxfile.attrib['version'] = "14.9.6"
#     mxfile.attrib['type'] = "device"
#     dwg = Drawing(mxfile)
#     if file_path is not None:
#         dwg.set_file_path(file_path)
#     return dwg

# def read(file_path):
#     '''
#         Read an existing draw.io diagram.

#         ----------

#         Arguments
#         -------------------------
#         `file_path` {str}
#             The file_path to the diagram.

#         Return {Drawing}
#         ----------------------
#         An instance of the Drawing class

#         Meta
#         ----------
#         `author`: Colemen Atwood
#         `created`: 05-27-2022 12:31:54
#         `memberOf`: drawing
#         `version`: 1.0
#         `method_name`: read
#     '''


#     xml = _f.readr(file_path)
#     # tree = etree.parse(xml)
#     tree = etree.fromstring(xml)
#     drawing = Drawing(tree)
#     drawing.data['file'] = _f.get_data(file_path)
#     return drawing

# class Drawing(_NodeBase):
#     '''
#         Used to manage drawio drawings and their diagrams.

#         ----------

#         Arguments
#         -------------------------
#         `tree` {etree}
#             A reference to the lxml etree

#         [`element`=None] {etree.element}
#             The drawing element of the xml file.
#             This is automatically generated by the new_drawing methods.

#             When a drawing is read, this class will automatically parse the drawing nodes.

#         Meta
#         ----------
#         `author`: Colemen Atwood
#         `created`: 06-05-2022 10:47:56
#         `memberOf`: drawing
#         `version`: 1.0
#         `name`: Drawing
#         * @xxx [06-05-2022 10:48:35]: documentation for Drawing
#     '''


#     def __init__(self,tree,element=None):
#         super().__init__(tree,element)
#         self.settings = {}
#         self.tree = tree
#         self.element = element
#         self.dia_root = None
#         self.data = {
#             "file":{
#                 "file_path":None,
#             },
#             "diagrams":[],
#         }

#         _get_diagrams(self)

#     def new_diagram(self,name):
#         '''
#             Create a new diagram in the drawing.

#             ----------

#             Arguments
#             -------------------------
#             `name` {str}
#                 The name of the diagram, this is shown on tabs in draw.io

#             Return {Diagram}
#             ----------------------
#             An instance of the Diagram class.

#             Meta
#             ----------
#             `author`: Colemen Atwood
#             `created`: 05-27-2022 12:34:35
#             `memberOf`: drawing
#             `version`: 1.0
#             `method_name`: new_diagram
#         '''

#         dia = _new_diagram(self.tree,name)
#         self.data['diagrams'].append(dia)
#         return dia

#     def save(self,path=None):
#         '''
#             Save this drawing

#             You should use set_file_path to set the default file_path for the drawing.

#             ----------

#             Arguments
#             -------------------------
#             [`path`=None] {str}
#                 Where to save this drawing, if not provided the default path will be used.

#             Keyword Arguments
#             -------------------------
#             `arg_name` {type}
#                     arg_description

#             Return {bool}
#             ----------------------
#             If no path is provided and there is no default path, it will return False.

#             Meta
#             ----------
#             `author`: Colemen Atwood
#             `created`: 05-27-2022 12:38:44
#             `memberOf`: drawing
#             `version`: 1.0
#             `method_name`: save
#         '''


#         # self.to_element()
#         if path is None:
#             if 'file_path' in self.data['file']:
#                 if self.data['file']['file_path'] is not None:
#                     path = self.data['file']
#             if path is None:
#                 return False
#         _f.write(path,etree.tostring(self.tree).decode("utf-8"))
#         return True

#     def list_diagrams(self):
#         '''
#             list all diagrams that belong to this drawing.

#             ----------

#             Return {None}
#             ----------------------
#             returns nothing

#             Meta
#             ----------
#             `author`: Colemen Atwood
#             `created`: 05-27-2022 12:40:32
#             `memberOf`: drawing
#             `version`: 1.0
#             `method_name`: list_diagrams
#         '''


#         for d in self.data['diagrams']:
#             if d.has_name():
#                 print(d.data['attributes']['name'])

#     def get_diagram(self,name:str) -> _diagram_type:
#         '''
#             Get a diagram by its name.

#             ----------

#             Arguments
#             -------------------------
#             `name` {str}
#                 The name of the diagram to return.

#             Return {Diagram}
#             ----------------------
#             The Diagram class instance with a matching name.

#             Meta
#             ----------
#             `author`: Colemen Atwood
#             `created`: 05-27-2022 12:41:56
#             `memberOf`: drawing
#             `version`: 1.0
#             `method_name`: get_diagram
#         '''
#         diagram:_diagram_type
#         for diagram in self.data['diagrams']:
#             if diagram.get_name(name):
#                 return diagram
#         return None

#     def set_file_path(self,value):
#         '''
#             Set the file_path for this drawing.

#             ----------

#             Arguments
#             -------------------------
#             `value` {str}
#                 The file_path to where this drawing will be saved.

#             Return {None}
#             ----------------------
#             returns nothing

#             Meta
#             ----------
#             `author`: Colemen Atwood
#             `created`: 05-27-2022 12:30:13
#             `memberOf`: drawing
#             `version`: 1.0
#             `method_name`: set_file_path
#         '''

#         self.data['file']['file_path'] = value

#     def get_file_path(self,test_value=None,default_val=None):
#         '''
#             Get the file_path attribute of the drawing.

#             ----------

#             Arguments
#             -------------------------
#             [`test_value`=None] {str}
#                 If provided, the file_path value must match this in order to return positively.

#             [`default_val`=''] {any}
#                 The value to return of the file_path does not exist or does not match the test_value.


#             Return {any}
#             ----------------------
#             If no test_value is provided the file_path value is returned.
#             If a test_value is provided and the file_path value matches, the file_path is returned.

#             If the file_path attribute does not exist or does not match the test_value,
#             the default_val is returned.

#             Meta
#             ----------
#             `author`: Colemen Atwood
#             `created`: 05-27-2022 11:55:25
#             `memberOf`: drawing
#             `version`: 1.0
#             `method_name`: get_file_path
#         '''

#         if 'file_path' in self.data['file']:
#             if test_value is not None:
#                 if self.data['file']['file_path'] == test_value:
#                     return self.data['file']['file_path']
#             else:
#                 return self.data['file']['file_path']
#         return default_val

# def _get_diagrams(drawing):
#     '''
#         Parses the drawing to find all diagrams that it contains.
#         This is used internally when reading a draw.io file.

#         ----------

#         Return {list}
#         ----------------------
#         A list of diagram instances, if no diagrams are found, the list is empty.

#         Meta
#         ----------
#         `author`: Colemen Atwood
#         `created`: 05-27-2022 12:35:48
#         `memberOf`: drawing
#         `version`: 1.0
#         `method_name`: get_diagrams
#     '''

#     result = []
#     diagrams = drawing.tree.xpath("//diagram['@name']")
#     if len(diagrams) > 0:
#         d:_diagram_type
#         for d in diagrams:
#             result.append(_Diagram(drawing.tree,d))
#     drawing.data['diagrams'] = result
#     return result


# # pylint: disable=missing-function-docstring
# # pylint: disable=missing-class-docstring
# # pylint: disable=line-too-long
# '''
#     Methods for managing drawio diagrams.

#     ----------

#     Meta
#     ----------
#     `author`: Colemen Atwood
#     `created`: 06-04-2022 15:45:00
#     `memberOf`: drawio
#     `name`: diagram
# '''

# from typing import Iterable,TypeVar
# from lxml import etree as _etree

# import utils.string_utils as _csu
# import utils.drawio.diagram_utils as _dia

# import utils.drawio.mcxell as _mxcell_module
# import utils.drawio.onode as _onode_module
# import utils.drawio.connector as connector_module
# import utils.drawio.drawing as _Drawing



# _new_mxcell = _mxcell_module.new_mxcell
# _Mxcell = _mxcell_module.Mxcell

# _Onode = _onode_module.Onode
# _new_onode = _onode_module.new_onode

# _Connector = connector_module.Connector
# _new_connector = connector_module.new_connector


# _onode_type = TypeVar('_onode_type', bound=_Onode)
# _connector_type = TypeVar('_connector_type', bound=_Connector)
# _mxcell_type = TypeVar('_mxcell_type', bound=_Mxcell)
# element_type = TypeVar('element_type', bound=_etree.Element)




# class Diagram:
#     '''
#         Manages a drawio diagram.
#         The diagrams contains the nodes and appear as tabs in the drawing.

#         ----------

#         Arguments
#         -------------------------
#         `tree` {etree}
#             The drawing xml tree.
#         `element` {element}
#             The etree element that will be added to the drawing.
#             The new_diagram method generates this automatically.

#         Meta
#         ----------
#         `author`: Colemen Atwood
#         `created`: 06-04-2022 16:47:17
#         `memberOf`: diagram
#         `version`: 1.0
#         `method_name`: Diagram
#         * @xxx [06-04-2022 16:49:56]: documentation for Diagram
#     '''


#     def __init__(self,tree:_etree,element=None):
#         self.settings = {}
#         self.tree = tree
#         self.element = element
#         self.dia_root = None
#         self.data = {
#             "attributes":{},
#             "connectors":[],
#             "mxcells":[],
#             "onodes":[],
#             "children":[],
#         }

#         self._from_element()
#         # self.set_defaults()

#     def _from_element(self):
#         element = self.element
#         if element is not None:
#             root_list:Iterable = _dia.get_diagram_root_node(element)
#             if len(root_list) > 0:
#                 dia_root = root_list[0]
#                 self.dia_root = dia_root
#             self.data['attributes'] = _dia.attrib_to_dict(element.attrib)
#             # self.data['connectors'] = _dia.get_connectors(element)
#             children = _dia.get_children(self.dia_root)
#             # print(f"children: {children}")
#             for c in children:
#                 if c.tag == "mxCell":
#                     if 'source' in c.attrib:
#                         con = _Connector(self.tree,c,self)
#                         self.data['connectors'].append(con)
#                         # self.data['connectors']
#                         # print(f"connector Found: {c}")
#                     else:
#                         self.data['mxcells'].append(_Mxcell(self.tree,c,self))
#                         # print(f"mxcell found: {c}")

#                 if c.tag == "object":
#                     O = _Onode(self.tree,c,self)
#                     self.data['onodes'].append(O)
#                     # print(f"object found: {c}")
#                 # self.data['children']['id'] = c.attrib['id']

#             return self.data

#     def show_grid(self,value:bool):
#         if value is True:
#             self.element.attrib['grid'] = "1"
#         if value is False:
#             self.element.attrib['grid'] = "0"

#     def grid_size(self,value:int):
#         if value < 1:
#             value = 10
#         self.element.attrib['gridSize'] = str(value)

#     def list_node_labels(self):
#         '''
#             Print the labels of every node in this diagram.

#             ----------


#             Return {None}
#             ----------------------
#             returns nothing.

#             Meta
#             ----------
#             `author`: Colemen Atwood
#             `created`: 05-27-2022 12:45:59
#             `memberOf`: diagram
#             `version`: 1.0
#             `method_name`: list_node_labels
#         '''
#         x:_Onode
#         for x in self.data['onodes']:
#             label = x.get_label(None)
#             if label is not None:
#                 print(label)

#         x:_Mxcell
#         for x in self.data['mxcells']:
#             label = x.get_label(None)
#             if label is not None:
#                 print(label)

#         x:_Connector
#         for x in self.data['connectors']:
#             label = x.get_label(None)
#             if label is not None:
#                 print(label)

#     def add_mxcell(self,id=None,parent=None)->_mxcell_type:
#         cell = _new_mxcell(self.tree,self,id,parent)
#         self.data['mxcells'].append(cell)
#         return cell

#     def add_onode(self,id=None)->_onode_type:
#         cell = _new_onode(self.tree,self,id)
#         self.data['onodes'].append(cell)
#         return cell

#     def add_connector(self,source,target)->_connector_type:
#         cell = _new_connector(self.tree,self,source,target)
#         self.data['connectors'].append(cell)
#         return cell

#     def get_nodes_by_tag(self,tag)->Iterable[_Onode]:
#         '''
#             Get all nodes that contain the tag provided.

#             ----------

#             Arguments
#             -------------------------
#             `tag` {str|list}
#                 The tag or list of tags to search for.


#             Return {list}
#             ----------------------
#             A list of onodes that contain the tag.
#             Only object nodes support the tags attribute, so this will not include mxcells or connectors.

#             Meta
#             ----------
#             `author`: Colemen Atwood
#             `created`: 05-27-2022 13:36:21
#             `memberOf`: diagram
#             `version`: 1.0
#             `method_name`: get_nodes_by_tag
#         '''

#         nodes = []
#         c:_Onode
#         for c in self.data['onodes']:
#             if c.has_tag(tag):
#                 nodes.append(c)
#         return nodes


#     def set_name(self,name:str):
#         '''
#             Set this diagrams name.

#             ----------

#             Arguments
#             -------------------------
#             `name` {str}
#                 The new name to assign.

#             Return {None}
#             ----------------------
#             returns nothing

#             Meta
#             ----------
#             `author`: Colemen Atwood
#             `created`: 05-27-2022 13:38:58
#             `memberOf`: diagram
#             `version`: 1.0
#             `method_name`: set_name
#         '''

#         self.data['attributes']['name'] = name
#         self.element.attrib['name'] = name

#     def get_name(self,test_value=None,default_val=False):
#         '''
#             Get the name attribute of the diagram.

#             ----------

#             Arguments
#             -------------------------
#             [`test_value`=None] {str}
#                 If provided, the name value must match this in order to return positively.

#             [`default_val`=''] {any}
#                 The value to return of the name does not exist or does not match the test_value.


#             Return {any}
#             ----------------------
#             If no test_value is provided the name value is returned.
#             If a test_value is provided and the name value matches, the name is returned.

#             If the name attribute does not exist or does not match the test_value,
#             the default_val is returned.

#             Meta
#             ----------
#             `author`: Colemen Atwood
#             `created`: 05-27-2022 11:55:25
#             `memberOf`: nodeBase
#             `version`: 1.0
#             `method_name`: get_name
#         '''

#         if 'name' in self.data['attributes']:
#             if test_value is not None:
#                 if self.data['attributes']['name'] == test_value:
#                     return self.data['attributes']['name']
#             else:
#                 return self.data['attributes']['name']
#         return default_val


#     def has_name(self,name=None):
#         if 'name' in self.data['attributes']:
#             if name is not None:
#                 if self.data['attributes']['name'] == name:
#                     return True
#             return True
#         return False


# def new_diagram(drawing:_Drawing,name:str)->Diagram:
#     '''
#         Create a new diagram in the drawing.

#         ----------

#         Arguments
#         -------------------------
#         `drawing` {Drawing}
#             A reference to the drawing that this diagram will belong to.
#         `name` {str}
#             The name of the diagram, this is shown on the tabs in draw.io

#         Return {Diagram}
#         ----------------------
#         An instance of the Diagram class

#         Meta
#         ----------
#         `author`: Colemen Atwood
#         `created`: 05-27-2022 12:42:40
#         `memberOf`: diagram
#         `version`: 1.0
#         `method_name`: new_diagram
#     '''

#     diagram = _etree.SubElement(drawing, 'diagram')
#     # diagram = _etree.Element("diagram")

#     diagram.attrib['id'] = _csu.gen.rand(20)
#     diagram.attrib['name'] = name

#     mxgm_data = {
#         "dx":"1074",
#         "dy":"954",
#         "grid":"1",
#         "gridSize":"10",
#         "guides":"1",
#         "tooltips":"1",
#         "connect":"1",
#         "arrows":"1",
#         "fold":"1",
#         "page":"1",
#         "pageScale":"1",
#         "pageWidth":"1700",
#         "pageHeight":"1100",
#         "math":"0",
#         "shadow":"0",
#     }

#     mxGraphModel = _etree.SubElement(diagram, 'mxGraphModel')
#     for k,v in mxgm_data.items():
#         mxGraphModel.attrib[k] = v

#     # @Mstep [] create the root node for the diagram
#     root = _etree.SubElement(mxGraphModel, 'root')


#     d = Diagram(drawing,diagram)

#     # @Mstep [] add the two default nodes to the diagram.
#     d.add_mxcell("0")
#     d.add_mxcell("1","0")

#     # @Mstep [RETURN] return the diagram instance.
#     return d






# pylint: disable=missing-function-docstring
# pylint: disable=missing-class-docstring
# pylint: disable=line-too-long
# pylint: disable=unused-import
'''
    A module of utility methods used manipulating lists.

    ----------

    Meta
    ----------
    `author`: Colemen Atwood
    `created`: 06-03-2022 10:22:15
    `memberOf`: list_utils
'''

import utils.dict_utils as _obj

from utils.string_utils.string_conversion import array_to_string_list as to_string_list
from utils.string_utils.string_format import array_replace_string as replace_from_list
from utils.string_utils.string_format import array_in_string as find_in_string



def append(base=None,value=None,**kwargs):
    '''
        Append an item to the base list.
        This is a lazy way of merging lists or appending a single item.

        ----------

        Arguments
        -------------------------
        `base` {list}
            The list to append an item to.
        `value` {any}
            The value to append to the base.

        Keyword Arguments
        -------------------------
        [`skip_null`=True] {bool}
            if True and the value is None, it will not append it.

        Return {type}
        ----------------------
        return_description

        Meta
        ----------
        `author`: Colemen Atwood
        `created`: 06-01-2022 08:45:33
        `memberOf`: objectUtils
        `version`: 1.0
        `method_name`: append
        # @TODO []: documentation for append
    '''

    if base is None:
        base = []

    skip_null = _obj.get_kwarg(["skip_null"],True,(bool),**kwargs)
    if skip_null is True:
        if value is None:
            return base

    if isinstance(value,(list)):
        base = base + value
    else:
        base.append(value)
    return base

def strip_list_nulls(value:list)->list:
    '''
        Strip None values from a list.

        ----------

        Arguments
        -------------------------
        `value` {list}
            The list to filter None values from.

        Return {list}
        ----------------------
        The list with all None values removed.

        Meta
        ----------
        `author`: Colemen Atwood
        `created`: 06-03-2022 08:38:50
        `memberOf`: objectUtils
        `version`: 1.0
        `method_name`: strip_list_nulls
        * @xxx [06-03-2022 08:39:37]: documentation for strip_list_nulls
    '''


    if isinstance(value,(list)) is False:
        return value
    return [x for x in value if x is not None]

def find_list_diff(list_one, list_two):
    '''
        find elements in list_one that do not exist in list_two.
        @param {list} list_one the primary list for comparison
        @param {list} list_two
        @function findListDiff
    '''
    return [x for x in list_one if x not in list_two]

def force_list(value)->list:
    '''
        Confirm that the value is a list, if not wrap it in a list.

        ----------

        Arguments
        -------------------------
        `value` {any}
            The value to test.

        Return {list}
        ----------------------
        The value as a list 

        Meta
        ----------
        `author`: Colemen Atwood
        `created`: 06-03-2022 09:13:57
        `memberOf`: object_utils
        `version`: 1.0
        `method_name`: force_list
        * @xxx [06-03-2022 09:14:52]: documentation for force_list
    '''


    if isinstance(value,(list)) is False:
        return [value]
    return value

def count(subj:list,value:any)->int:
    '''
        Count how many times a value occurs in a list.
        This is case sensitive, it will only count exact matches.

        ----------

        Arguments
        -------------------------
        `subj` {list}
            The list to search.

        `value` {any}
            The value to search for.

        Return {int}
        ----------------------
        The number of times the value occurs in the list.

        Meta
        ----------
        `author`: Colemen Atwood
        `created`: 06-07-2022 15:43:26
        `memberOf`: list_utils
        `version`: 1.0
        `method_name`: count
        * @xxx [06-07-2022 15:44:46]: documentation for count
    '''


    return len([x for x in subj if x == value])









# pylint: disable=missing-function-docstring
# pylint: disable=missing-class-docstring
# pylint: disable=line-too-long
# pylint: disable=line-too-long
# pylint: disable=unused-import

'''
    A module of utility methods used for parsing strings

    ----------

    Meta
    ----------
    `author`: Colemen Atwood
    `created`: 06-04-2022 10:44:23
    `memberOf`: string_utils
'''


# import json
# import hashlib
# import string
import re
import utils.sql_utils as sql

def determine_gender(value:str)->str:
    '''
        Use synonyms to determine the gender of a word.

        ----------

        Arguments
        -------------------------
        `value` {str}
            The word to test

        Return {str|None}
        ----------------------
        Either "male", "female" or None

        Meta
        ----------
        `author`: Colemen Atwood
        `created`: 06-04-2022 09:00:38
        `memberOf`: string_parsing
        `version`: 1.0
        `method_name`: determine_gender
        * @xxx [06-04-2022 09:02:44]: documentation for determine_gender
    '''


    female_synonyms = ['girl','female','woman','lady','miss','chica','lass','chick','grandmother','grandma','mom','mother','daughter','wife']
    male_synonyms = ['boy','male','man','dude','guy','husband','bro','grandfather','grandpa','dad','father','brother']

    value = value.lower()
    if value in female_synonyms:
        return "female"
    if value in male_synonyms:
        return "male"
    return None



# pylint: disable=missing-function-docstring
# pylint: disable=missing-class-docstring
# pylint: disable=line-too-long
'''
    A module of utility methods used for manipulating dictionaries.

    ----------

    Meta
    ----------
    `author`: Colemen Atwood
    `created`: 06-03-2022 10:22:15
    `memberOf`: dict_utils
'''
# import random

from typing import Union as _Union
from colorama import Fore as _Fore
from colorama import Style as _Style
import utils.list_utils as _lu

def set_defaults(default_vals, obj,**kwargs):
    '''
        Sets default values on the dict provided, if they do not already exist or 
        if the value is None.

        ----------

        Arguments
        -------------------------
        `default_vals` {dict}
            The default values to set on the obj.
        `obj` {dict}
            The object to assign default values to.

        Keyword Arguments
        -------------------------
        [`replace_null`=False] {bool}
            If True, None values in the obj dict are overwritten by the defaults.

        Return {dict}
        ----------------------
        The obj with default values applied

        Meta
        ----------
        `author`: Colemen Atwood
        `created`: 12-09-2021 08:04:03
        `memberOf`: object_utils
        `version`: 1.0
        `method_name`: set_defaults
    '''
    replace_null = get_kwarg(['replace null'],False,(bool),**kwargs)
    
    for k, v in default_vals.items():
        if replace_null:
            if k in obj:
                if obj[k] is None:
                    obj[k] = v

        if k not in obj:
            obj[k] = v
        # print(f"k: {k} - v: {v}")
    return obj

def merge(dict_one:dict,dict_two:dict)->dict:
    '''
        Merge two dictionaries into one.

        ----------

        Arguments
        -------------------------
        `dict_one` {dict}
            The first dict to merge.
        `dict_two` {dict}
            The second dict to merge.

        Return {dict}
        ----------------------
        The merged dictionary

        Meta
        ----------
        `author`: Colemen Atwood
        `created`: 06-05-2022 14:19:19
        `memberOf`: dict_utils
        `version`: 1.0
        `method_name`: merge
        * @xxx [06-05-2022 14:20:44]: documentation for merge
    '''
    return {**dict_one,**dict_two}

def keys_to_lower(dictionary):
    '''
        Converts all keys in a dictionary to lowercase.
    '''
    return {k.lower(): v for k, v in dictionary.items()}

def keys_to_list(data:dict)->list:
    '''
        return all keys in a dictionary as a list.

        ----------

        Arguments
        -------------------------
        `data` {dict}
            The dictionary to parse.

        Return {list|None}
        ----------------------
        A list of the keys in the dictionary.
        returns an empty list if it fails or a non-dictionary was provided.

        Meta
        ----------
        `author`: Colemen Atwood
        `created`: 06-03-2022 07:49:21
        `memberOf`: objectUtils
        `version`: 1.0
        `method_name`: keys_to_list
        * @xxx [06-03-2022 07:50:27]: documentation for keys_to_list
    '''
    if isinstance(data,(dict)) is False:
        return []

    return list(data.keys())

def get_kwarg(key_name:_Union[list,str], default_val=False, value_type=None, **kwargs):
    '''
        Get a kwarg argument that optionally matches a type check or
        return the default value.

        ----------

        Arguments
        -------------------------
        `key_name` {list|string}
            The key name or a list of key names to search kwargs for.

        [`default_val`=False] {any}
            The default value to return if the key is not found or fails
            the type check (if provided.)

        [`value_type`=None] {any}
            The type or tuple of types.
            The kwarg value must match at least one of these.
            Leave as None to ignore type checking.
        `kwargs` {dict}
            The kwargs dictionary to search within.

        Return {any}
        ----------------------
        The value of the kwarg key if it is found.
        The default value if the key is not found or its value fails type checking.

        Meta
        ----------
        `author`: Colemen Atwood
        `created`: 06-03-2022 08:33:36
        `memberOf`: objectUtils
        `version`: 1.0
        `method_name`: get_kwarg
        * @xxx [06-03-2022 08:38:33]: documentation for get_kwarg
    '''
    from utils.random_utils.rand_generation import gen_variations


    kwargs = keys_to_lower(kwargs)
    if isinstance(key_name, list) is False:
        key_name = [key_name]

    for name in key_name:
        # generate basic variations of the name
        varis = gen_variations(name)
        for v_name in varis:
            if v_name in kwargs:
                if value_type is not None:
                    if isinstance(kwargs[v_name], value_type) is True:
                        return kwargs[v_name]
                else:
                    return kwargs[v_name]
    return default_val

def get_arg(args:dict,key_name:_Union[list,str],default_val=False, value_type=None)->any:
    '''
        Get a key's value from a dictionary.

        ----------

        Arguments
        -------------------------
        `args` {dict}
            The dictionary to search within.

        `key_name` {str|list}
            The key or list of keys to search for.

        [`default_val`=False] {any}
            The value to return if the key is not found.

        [`value_type`=None] {any}
            The type the value should have. This can be a tuple of types.

        Return {any}
        ----------------------
        The key's value if it is found and matches the value_type (if provided.)
        The default value otherwise.

        Meta
        ----------
        `author`: Colemen Atwood
        `created`: 06-02-2022 07:43:12
        `memberOf`: object_utils
        `version`: 1.0
        `method_name`: get_arg
        * @xxx [06-02-2022 07:46:35]: documentation for get_arg
    '''
    from utils.random_utils.rand_generation import gen_variations

    if isinstance(args,(dict)) is False:
        return default_val
    if len(args.keys()) == 0:
        return default_val

    args = keys_to_lower(args)
    # if defaults is not None:
    #     defaults = keys_to_lower(defaults)
    #     args = set_defaults(defaults,args)

    if isinstance(key_name, list) is False:
        key_name = [key_name]

    for name in key_name:
        # generate basic variations of the name
        varis = gen_variations(name)
        for v_name in varis:
            if v_name in args:
                if value_type is not None:
                    if isinstance(args[v_name], value_type) is True:
                        return args[v_name]
                else:
                    return args[v_name]
    return default_val

def get_unique_keys(obj, **kwargs):
    '''
        Gets all unique keys in the object provided.

        @param {dict|list} obj - The object or list to search for keys within.
        @param {boolean} [**sort_list=True] - Sort the list alphabetically.
        @param {boolean} [**case_sensitive=True] - If True the case of the key is ignored.
        @param {boolean} [**force_lowercase=True] - Convert all keys to lowercase.
        @param {boolean} [**recursive=True] - Recurse into nested objects to find keys.
        @param {int} [**max_depth=500] - The maximum recursions it is allowed to make.
        @return {list} A list of unique keys from the object, if none are found the list is empty.
        @function get_unique_keys
    '''

    __current_depth = get_kwarg(['__current_depth'], 0, int, **kwargs)
    sort_list = get_kwarg(['sort_list'], False, bool, **kwargs)
    case_sensitive = get_kwarg(['case_sensitive'], True, bool, **kwargs)
    force_lowercase = get_kwarg(['force_lowercase'], True, bool, **kwargs)
    recursive = get_kwarg(['recursive'], True, bool, **kwargs)
    max_depth = get_kwarg(['max_depth'], 500, int, **kwargs)
    kwargs['__current_depth'] = __current_depth + 1

    keys = []

    if recursive is True and __current_depth < max_depth:
        if isinstance(obj, (list, tuple, set)):
            for element in obj:
                if isinstance(element, (list, dict)):
                    keys = keys + get_unique_keys(element, **kwargs)

    if isinstance(obj, dict):
        keys = list(obj.keys())

        if recursive is True and __current_depth < max_depth:
            # pylint: disable=unused-variable
            for k, value in obj.items():
                # find nested objects
                if isinstance(value, (list, dict, tuple, set)):
                    keys = keys + get_unique_keys(value, **kwargs)

    if case_sensitive is True:
        output = []
        lkkeys = []
        for key in keys:
            low_key = key.lower()
            if low_key not in lkkeys:
                output.append(key)
                lkkeys.append(low_key)
        keys = output

    if force_lowercase is True:
        keys = [x.lower() for x in keys]

    keys = list(set(keys))

    if sort_list is True:
        keys = sorted(keys, key=lambda x: int("".join([i for i in x if i.isdigit()])))
    return keys

def has_keys(data:dict,keys:list,**kwargs):
    '''
        confirm that a dictionary has all keys in the key list.

        ----------

        Arguments
        -------------------------
        `data` {dict}
            The dictionary to validate.

        `keys` {list}
            A list of keys that the data dict must contain.

        Keyword Arguments
        -------------------------
        [`message_template`=None] {str}
            The message to print to the console log if a key is missing.
            The string __KEY__ will be replaced with the missing key name.

        Return {bool}
        ----------------------
        True if the dict contains all the keys, False otherwise.

        Meta
        ----------
        `author`: Colemen Atwood
        `created`: 06-03-2022 09:15:17
        `memberOf`: object_utils
        `version`: 1.0
        `method_name`: has_keys
        * @xxx [06-03-2022 09:18:47]: documentation for has_keys
    '''


    message_template = get_kwarg(['message_template'], None, (str), **kwargs)
    missing_keys = []
    keys = _lu.force_list(keys)
    for k in keys:
        if k not in data:
            if message_template is not None:
                msg = message_template.replace("__KEY__",k)
                print(_Fore.RED + msg + _Style.RESET_ALL)
            missing_keys.append(k)
    if len(missing_keys) > 0:
        return False
    return True

def remove_keys(data:dict,keys:_Union[list,str],**kwargs)->dict:
    '''
        Remove matching keys from a dictionary or keep only the matching keys.

        ----------

        Arguments
        -------------------------
        `data` {dict}
            The dictionary to filter.

        `keys` {list|str}
            A key or list of keys that will be removed from the dictionary.

        Keyword Arguments
        -------------------------
        [`reverse`=False] {bool}
            If True, all keys except the ones provided will be removed.

        Return {dict}
        ----------------------
        The dict with keys filtered.

        Meta
        ----------
        `author`: Colemen Atwood
        `created`: 06-04-2022 10:15:45
        `memberOf`: dict_utils
        `version`: 1.0
        `method_name`: remove_keys
        * @xxx [06-04-2022 10:23:17]: documentation for remove_keys
    '''
    reverse = get_kwarg(['reverse'], False, (bool), **kwargs)
    strip_nulls = get_kwarg(['strip_nulls'], False, (bool), **kwargs)
    keys = _lu.force_list(keys)


    output = {}
    for k,v in data.items():
        if strip_nulls is True:
            if v is None:
                continue
        if reverse is True:
            if k in keys:
                output[k] = v
        else:
            if k not in keys:
                output[k] = v

    return output

def strip_nulls(data:dict)->dict:
    '''
        Remove all keys with a None value in the dictionary.

        ----------

        Arguments
        -------------------------
        `data` {dict}
            The dictionary to filter.

        Keyword Arguments
        -------------------------
        `arg_name` {type}
                arg_description

        Return {type}
        ----------------------
        return_description

        Meta
        ----------
        `author`: Colemen Atwood
        `created`: 06-08-2022 08:10:41
        `memberOf`: dict_utils
        `version`: 1.0
        `method_name`: def strip_nulls(data:dict)->dict:
        * @TODO []: documentation for def strip_nulls(data:dict)->dict:
    '''


    new_data = {}
    for k,v in data.items():
        if v is not None:
            new_data[k] = v
    return new_data



# pylint: disable=missing-function-docstring
# pylint: disable=missing-class-docstring
# pylint: disable=line-too-long
# pylint: disable=unused-wildcard-import
# pylint: disable=wildcard-import
# pylint: disable=unused-import
'''
    A module of utility methods used for parsing SQL code.

    ----------

    Meta
    ----------
    `author`: Colemen Atwood
    `created`: 06-03-2022 10:04:06
    `memberOf`: parse_sql
    `version`: 1.0
    `method_name`: parse_sql
'''


import re as _re
import shlex as _shlex
import sqlparse as _sqlparse
from typing import Union as _Union

from pyparsing import And, Suppress,Dict, Word, Literal, Group, Optional, ZeroOrMore, OneOrMore, Regex, restOfLine, alphanums,nums, printables, string, CaselessKeyword,nestedExpr,ParseException,quotedString,removeQuotes,originalTextFor,delimitedList,QuotedString

import utils.dict_utils as _obj
import utils.list_utils as _lu
import utils.string_utils as _csu
import utils.file_utils as _f
import utils.sql_utils as _sql
import config as _config
_log = _config.log

_sql_data_types = [
    "mediumtext",
    "mediumblob",
    "varbinary",
    "timestamp",
    "mediumint",
    "tinytext",
    "tinyblob",
    "smallint",
    "longtext",
    "longblob",
    "datetime",
    "varchar",
    "tinyint",
    "integer",
    "decimal",
    "boolean",
    "double",
    "double",
    "binary",
    "bigint",
    "float",
    "float",
    "year",
    "time",
    "text",
    "enum",
    "date",
    "char",
    "bool",
    "blob",
    "set",
    "int",
    "dec",
    "bit",
]

_sql_types_with_sizes = [
    "varbinary",
    "mediumint",
    "timestamp",
    "smallint",
    "datetime",
    "varchar",
    "decimal",
    "integer",
    "tinyint",
    "binary",
    "double",
    "double",
    "bigint",
    "float",
    "float",
    "text",
    "char",
    "time",
    "blob"
    "dec",
    "int",
    "bit",
]

def sql_type_to_python(value):
    '''
        Attempts to convert an sql type to its python equivalent.

        Keep in mind this is not exact.. at fucking all..
        This is attempting to predict how the value will need to be stored in the SQL file.

        ----------

        Arguments
        -------------------------
        `value` {str}
            The SQL type to convert.

        Return {list}
        ----------------------
        A list of similar types

        Meta
        ----------
        `author`: Colemen Atwood
        `created`: 06-02-2022 07:20:35
        `memberOf`: parse_sql
        `version`: 1.0
        `method_name`: sql_type_to_python
        * @xxx [06-02-2022 07:22:30]: documentation for sql_type_to_python
    '''

    if value in _sql_data_types:
        return _sql_data_types[value]
    return None

def parse_column_data(value=None):

    # value = "`timestamp`       int NULL DEFAULT NULL COMMENT 'Timestamp of when the task was created.' ,"
    _tick,_double_quote,_single_quote,_comma = [Suppress(x) for x in list('`"\',')]
    _quote = (_tick|_double_quote|_single_quote)
    _open_paren,_close_paren = [Suppress(x) for x in list('()')]
    # paren_word = _open_paren + Optional(_quote) + Word(alphanums + "_") + Optional(_quote) + _close_paren
    quoted_word = _quote + Word(alphanums + "_") + _quote


    column_name = quoted_word.setResultsName('name')


    ctypes = CaselessKeyword("mediumtext") | CaselessKeyword("mediumblob") | CaselessKeyword("varbinary") | CaselessKeyword("timestamp") | CaselessKeyword("mediumint") | CaselessKeyword("tinytext") | CaselessKeyword("tinyblob") | CaselessKeyword("smallint") | CaselessKeyword("longtext") | CaselessKeyword("longblob") | CaselessKeyword("datetime") | CaselessKeyword("varchar") | CaselessKeyword("tinyint") | CaselessKeyword("integer") | CaselessKeyword("decimal") | CaselessKeyword("boolean") | CaselessKeyword("double") | CaselessKeyword("double") | CaselessKeyword("binary") | CaselessKeyword("bigint") | CaselessKeyword("float") | CaselessKeyword("float") | CaselessKeyword("year") | CaselessKeyword("time") | CaselessKeyword("text") | CaselessKeyword("enum") | CaselessKeyword("date") | CaselessKeyword("char") | CaselessKeyword("bool") | CaselessKeyword("blob") | CaselessKeyword("set") | CaselessKeyword("int") | CaselessKeyword("dec") | CaselessKeyword("bit")
    ctype_len = Optional(_open_paren + Word(nums) + _close_paren)
    column_type = ctypes.setResultsName('type') + ctype_len.setResultsName('size')

    null_vals = CaselessKeyword("NULL") | CaselessKeyword("NOT NULL")
    null_type = null_vals.setResultsName('null_type')

    def_k = Suppress(CaselessKeyword('DEFAULT'))
    default = Optional(def_k + Word(alphanums))
    quote = QuotedString("'", esc_quote="''",unquote_results=False).search_string(value)
    if len(quote) > 0:
        quote = quote[0][0]
        default = Optional(default|def_k + quote).setResultsName('default')

    com_k = Suppress(CaselessKeyword('COMMENT'))
    comment = Optional(com_k + Word(alphanums))
    quote = QuotedString("'", esc_quote="''",unquote_results=False).search_string(value)
    if len(quote) > 0:
        quote = quote[0][0]
        comment = Optional(com_k + quote).setResultsName('comment')



    grammar = column_name + column_type + null_type + default + comment
    res = ZeroOrMore(Group(grammar).setResultsName('data')).parseString(value)

    output = None

    if len(res) > 0:
        data = res.as_dict()['data']
        # print(data)
        defaults = {
            "name":None,
            "type":None,
            "allow_nulls":None,
            "default":None,
            "comment":None,
        }
        new_data = {}
        for k,v in data.items():
            if k == "size":
                new_data[k] = int(v[0])
                continue
            if k == "null_type":
                if v.upper() == 'NOT NULL':
                    new_data['allow_nulls'] = False
                    continue
                if v.upper() == 'NULL':
                    new_data['allow_nulls'] = True
                    continue
            if isinstance(v,(list)):
                v = _csu.strip(v[0],["'",'"'])
                new_data[k] = v
                continue
            new_data[k] = v

        # print(new_data)
        new_data = _obj.set_defaults(defaults,new_data)
        output = new_data
    return output

def parse_schema_statement(value=None):
    
    # value = "CREATE SCHEMA IF NOT EXISTS `idealech_Equari_Content_Database`;"


    _tick,_double_quote,_single_quote,_period = [Suppress(x) for x in list('`"\'.')]
    _quote = (_tick|_double_quote|_single_quote)
    _open_paren,_close_paren = [Suppress(x) for x in list('()')]
    # paren_word = _open_paren + Optional(_quote) + Word(alphanums + "_") + Optional(_quote) + _close_paren
    quoted_word = _quote + Word(alphanums + "_") + _quote

    keys = CaselessKeyword("schema")
    create_statement = CaselessKeyword("create").setResultsName('action') + keys
    drop_statement = CaselessKeyword("drop").setResultsName('action') + keys
    action = drop_statement | create_statement


    exists = Optional(CaselessKeyword("if exists") | CaselessKeyword("if not exists")).setResultsName('test')
    schema_name = quoted_word.setResultsName('schema_name')


    grammar = action + exists + schema_name
    res = ZeroOrMore(Group(grammar).setResultsName('data')).parseString(value)

    output = None

    if len(res) > 0:
        data = res.as_dict()['data']
        # print(data)
        defaults = {
            "action":None,
            "test":None,
            "schema_name":None,
            "raw_statement":value,
        }
        new_data = {}
        for k,v in data.items():
            if isinstance(v,(list)):
                new_data[k] = v[0]
                continue
            new_data[k] = v
        new_data = _obj.set_defaults(defaults,new_data)
        # print(new_data)
        output = new_data
    return output

#/ =================================================================================================
def parse_table_statement(value:_Union[None,str]=None)->_Union[dict,None]:
    '''
        Parse a table SQL statement.

        ----------

        Arguments
        -------------------------
        `value` {str,None}
            The SQL string to parse.

        Return {dict,None}
        ----------------------
        The statement data dictionary if it can be parsed, None otherwise.
        
        {
            `action`: "create",
            `test`: "if not exists",
            `schema_name`: "idealech_Equari_Management_Database",
            `table_name`: "faq_categories",
            `raw_statement`: "Original and Unmodified Create Statement Goes here.",
            `columns`: [
                {
                    "name": "faq_category_id",
                    "type": "int",
                    "allow_nulls": false,
                    "default": null,
                    "comment": null,
                    "is_primary_key": false,
                    "primary_key": true
                },
                ...
            ],
            `primary_keys`: [
                "faq_category_id"
            ],
            `keys`: [
                {
                    "key_name": "FK_FaqCategories_FaqCategoryID",
                    "foreign_col_name": [
                        "parent_faq_category_id"
                    ]
                }
            ],
            `constraints`: [
                {
                    "constraint_name": "FK_FaqCategory_Parent_FaqCategory",
                    "foreign_key": "FK_FaqCategories_FaqCategoryID",
                    "local_col_name": "parent_faq_category_id",
                    "schema_name": "idealech_Equari_Management_Database",
                    "table_name": "faq_categories",
                    "foreign_col_name": "faq_category_id",
                    "on_delete": "CASCADE",
                    "on_update": "CASCADE"
                }
            ],
            `name`: "faq_categories"
        }



        Meta
        ----------
        `author`: Colemen Atwood
        `created`: 06-05-2022 20:30:22
        `memberOf`: sql_parse
        `version`: 1.0
        `method_name`: parse_table_statement
        * @xxx [06-05-2022 20:34:11]: documentation for parse_table_statement
    '''
    # _log("sql_parse.parse_table_statement")

    _tick,_double_quote,_single_quote,_period = [Suppress(x) for x in list('`"\'.')]
    _quote = (_tick|_double_quote|_single_quote)
    # _open_paren,_close_paren = [Suppress(x) for x in list('()')]
    # paren_word = _open_paren + Optional(_quote) + Word(alphanums + "_") + Optional(_quote) + _close_paren
    quoted_word = _quote + Word(alphanums + "_") + _quote


    keys = CaselessKeyword("table")
    create_statement = CaselessKeyword("create").setResultsName('action') + keys
    drop_statement = CaselessKeyword("drop").setResultsName('action') + keys
    action = drop_statement | create_statement


    exists = Optional(CaselessKeyword("if exists") | CaselessKeyword("if not exists")).setResultsName('test')
    table_name = Optional(quoted_word.setResultsName('schema_name') + _period) + quoted_word.setResultsName('table_name')


    grammar = action + exists + table_name
    res = ZeroOrMore(Group(grammar).setResultsName('data')).parseString(value)

    output = None

    if len(res) > 0:
        data = res.as_dict()['data']
        defaults = {
            "raw_statement":value,
            "action":None,
            "test":None,
            "schema_name":None,
            "table_name":None,
        }
        new_data = {}
        for k,v in data.items():
            if isinstance(v,(list)):
                new_data[k] = v[0]
                continue
            new_data[k] = v
        new_data = _obj.set_defaults(defaults,new_data)
        if new_data['action'] == "create":
            cols = capture_create_table_columns(value)
            # new_data['column_data'] = _obj.strip_list_nulls([parse_column_data(x) for x in cols])
            d = _parse_table_column_lines(cols)
            new_data["columns"] = d['columns']
            new_data["primary_keys"] = d['primary_keys']
            new_data["keys"] = d['keys']
            new_data["constraints"] = d['constraints']
        # print(new_data)
        output = new_data

    return output

def _parse_table_column_lines(lines):
    data = {
        "columns":[],
        "primary_keys":[],
        "keys":[],
        "constraints":[],
    }
    # keys = []

    for line in lines:
        # print(f"line: {line}")
        if is_line_key(line):
            # print(f"key found: {line} {parse_key(line)}")
            data['keys'] = _lu.append(data['keys'],parse_key(line))
            pk = parse_primary_key(line)
            if pk is not None:
                data['primary_keys'] = _lu.append(data['primary_keys'],pk['primary_key'])
            # data['primary_keys'] = _lu.append(data['primary_keys'],parse_primary_key(line))

        if is_line_constraint(line):
            data['constraints'] = _lu.append(data['constraints'],parse_constraint(line))


        data['columns'] = _lu.append(data['columns'],parse_column_data(line))

        # parse_column_data(line)
    # data['keys'] = keys
    for c in data['columns']:
        c['is_primary_key'] = False
        if c['name'] in data['primary_keys']:
            c['primary_key'] = True
    return data

def capture_create_table_columns(sql):
    '''
        Used by parse_table_statement to capture the column area of the statement.

        ----------

        Arguments
        -------------------------
        `sql` {str}
            The create table statement to parse.

        Return {list}
        ----------------------
        A list of column declarations upon success.
        The list is empty if nothing is found.

        Meta
        ----------
        `author`: Colemen Atwood
        `created`: 06-01-2022 08:48:11
        `memberOf`: parse_sql
        `version`: 1.0
        `method_name`: capture_create_table_columns
        @xxx [06-01-2022 08:50:01]: documentation for capture_create_table_columns
    '''

    output = []
    scanner = originalTextFor(nestedExpr('(',')'))
    for match in scanner.searchString(sql):
        val = _csu.strip(match[0],["(",")"])
        val = _sql.escape_quoted_chars(val,True)
        output = val.split("\n")
    return output


def _parse_list(value:str)->str:
    '''
        A utility function used for parsing an SQL string list into a more formatted version.

        ----------

        Arguments
        -------------------------
        `arg_name` {type}
                arg_description

        Keyword Arguments
        -------------------------
        `arg_name` {type}
                arg_description

        Return {type}
        ----------------------
        return_description

        Meta
        ----------
        `author`: Colemen Atwood
        `created`: 06-05-2022 20:27:02
        `memberOf`: sql_parse
        `version`: 1.0
        `method_name`: _parse_list
        * @xxx [06-05-2022 20:28:56]: documentation for _parse_list
    '''


    value = value.asList()
    value = _sql.strip_sql_quotes(value[0])
    vals = value.split(",")
    vals = [_csu.strip(x,[" "]) for x in vals]
    return ','.join(vals)

def parse_key(value=None):
    if 'key' not in value.lower():
        return None
    # value ="KEY `FK_ActivityLogs_ActivityTypeID` (`activity_type_id`, `activity_type_hash_id`),"
    _tick,_double_quote,_single_quote,_comma = [Suppress(x) for x in list('`"\',')]
    _quote = (_tick|_double_quote|_single_quote)
    _open_paren,_close_paren = [Suppress(x) for x in list('()')]
    # paren_word = _open_paren + Optional(_quote) + Word(alphanums + "_") + Optional(_quote) + _close_paren
    quoted_word = _quote + Word(alphanums + "_") + _quote

    paren_word_list = _open_paren + Word(alphanums + "_,'\"` ") + _close_paren

    # KEY `FK_ActivityLogs_RequestLogID`
    key_name = Suppress(CaselessKeyword('KEY')) + quoted_word.setResultsName('key_name')

    foreign_col_name = paren_word_list.setResultsName('foreign_col_name').setParseAction(_parse_list)

    grammar = key_name + foreign_col_name
    res = ZeroOrMore(Group(grammar).setResultsName('data')).parseString(value)

    output = None

    if len(res) > 0:
        data = res.as_dict()['data']
        new_data = {}
        for k,v in data.items():
            if k == "foreign_col_name":
                new_data[k] = v.split(",")
                continue
            if isinstance(v,(list)):
                new_data[k] = v[0]
                continue
            new_data[k] = v

        # print(new_data)
        output = new_data
    return output

def parse_primary_key(value=None):
    if 'primary key' not in value.lower():
        return None
    # value = "PRIMARY KEY (`activity_log_id`)"
    # value = "PRIMARY KEY (`activity_type_id`, `activity_type_hash_id`),"
    _tick,_double_quote,_single_quote,_period = [Suppress(x) for x in list('`"\'.')]
    _quote = (_tick|_double_quote|_single_quote)
    _open_paren,_close_paren = [Suppress(x) for x in list('()')]
    # paren_word = _open_paren + Optional(_quote) + Word(alphanums + "_") + Optional(_quote) + _close_paren
    # quoted_word = _quote + Word(alphanums + "_") + _quote
    paren_word_list = _open_paren + Word(alphanums + "_,'\"` ") + _close_paren
    primary_key = paren_word_list.setResultsName('primary_key').setParseAction(_parse_list)
    grammar = Suppress(CaselessKeyword('primary key')) + primary_key
    res = ZeroOrMore(Group(grammar).setResultsName('data')).parseString(value)

    output = None

    if len(res) > 0:
        data = res.as_dict()['data']
        new_data = {}
        for k,v in data.items():
            if k == "primary_key":
                new_data[k] = v.split(",")
                continue
            if isinstance(v,(list)):
                new_data[k] = v[0]
                continue
            new_data[k] = v

        # print(new_data)
        output = new_data
    return output

def parse_constraint(value:str):
    '''
        Parse constraint data from the string.

        ----------

        Arguments
        -------------------------
        `value` {str}
            The string to parse.

        Return {dict,None}
        ----------------------
        The constraint data dictionary if successful, None otherwise.

        Meta
        ----------
        `author`: Colemen Atwood
        `created`: 06-05-2022 20:20:57
        `memberOf`: sql_parse
        `version`: 1.0
        `method_name`: parse_constraint
        * @xxx [06-05-2022 20:26:05]: documentation for parse_constraint
    '''


    _tick,_double_quote,_single_quote,_period = [Suppress(x) for x in list('`"\'.')]
    _quote = (_tick|_double_quote|_single_quote)
    _open_paren,_close_paren = [Suppress(x) for x in list('()')]
    paren_word = _open_paren + Optional(_quote) + Word(alphanums + "_") + Optional(_quote) + _close_paren
    quoted_word = _quote + Word(alphanums + "_") + _quote


    constraint_conditions = (CaselessKeyword('RESTRICT') | CaselessKeyword('CASCADE') | CaselessKeyword('SET NULL') | CaselessKeyword('NO ACTION') | CaselessKeyword('SET DEFAULT'))


    constraint_name = Suppress(CaselessKeyword('CONSTRAINT')) + quoted_word.setResultsName('constraint_name')


    foreign_key_name = Suppress(CaselessKeyword('FOREIGN KEY')) + quoted_word.setResultsName('foreign_key')
    local_col_name = paren_word.setResultsName('local_col_name')
    foreign_table = Suppress(CaselessKeyword('REFERENCES')) + Optional(quoted_word.setResultsName('schema_name') + _period) + quoted_word.setResultsName('table_name')
    foreign_col_name = paren_word.setResultsName('foreign_col_name')

    on_delete = Suppress(CaselessKeyword('ON DELETE')) + constraint_conditions.setResultsName('on_delete')
    on_update = Suppress(CaselessKeyword('ON UPDATE')) + constraint_conditions.setResultsName('on_update')
    constraints = ZeroOrMore(on_delete | on_update)

    grammar = constraint_name + foreign_key_name + local_col_name + foreign_table + foreign_col_name + constraints
    res = ZeroOrMore(Group(grammar).setResultsName('data')).parseString(value)

    output = None

    if len(res) > 0:
        data = res.as_dict()['data']
        new_data = {}
        for k,v in data.items():
            if isinstance(v,(list)):
                new_data[k] = v[0]
                continue
            new_data[k] = v

        # print(new_data)
        output = new_data
    return output

def parse_table_columns(sql):
    '''
        Parses an SQL create file into a list of column dictionaries.

        ----------

        Arguments
        -------------------------
        `sql` {str}
            The sql to parse or a file path to parse.

        Keyword Arguments
        -------------------------
        `arg_name` {type}
                arg_description

        Return {type}
        ----------------------
        return_description

        Meta
        ----------
        `author`: Colemen Atwood
        `created`: 05-31-2022 07:59:37
        `memberOf`: TableManager
        `version`: 1.0
        `method_name`: parse_table_columns
        # @TODO []: documentation for parse_table_columns
    '''

    if _f.exists(sql):
        sql = _f.readr(sql)
    file_data = strip_keys_constraints(sql)
    file_data['columns'] = {}

    cols = {}
    parsed = _sqlparse.parse(file_data['sql'])
    for stmt in parsed:
        # Get all the tokens except whitespaces
        tokens = [t for t in _sqlparse.sql.TokenList(stmt.tokens) if t.ttype != _sqlparse.tokens.Whitespace]
        is_create_stmt = False
        for _, token in enumerate(tokens):
            # Is it a create statements ?
            if token.match(_sqlparse.tokens.DDL, 'CREATE'):
                is_create_stmt = True
                continue

            # If it was a create statement and the current token starts with "("
            if is_create_stmt and token.value.startswith("("):
                # Get the table name by looking at the tokens in reverse order till you find
                # a token with None type
                # print (f"table: {get_table_name(tokens[:i])}")

                # Now parse the columns
                txt = token.value
                # strip_indices(txt)
                # print("\n\n")
                columns = txt[1:txt.rfind(")")].replace("\n","").split(",")
                # print(columns)
                # print("\n\n")
                for column in columns:
                    c = ' '.join(column.split()).split()
                    # print(c)
                    if len(c) == 0:
                        continue
                    c_name = c[0].replace('\"',"")
                    c_name = c_name.replace('`','')
                    # c_type = c[1]  # For condensed type information
                    # OR
                    c_type = " ".join(c[1:]) # For detailed type informatio
                    data = {
                        "name":c_name,
                        "raw_type":c_type,
                        "default":None,
                        "size":None,
                        "allow_nulls":False,
                        "data_type":None,
                        "comment":None,
                        "shlexed":None,
                        "is_primary_key":False,
                    }
                    # print (f"raw_type: {data['raw_type']}")
                    data['shlexed'] = _shlex.split(data['raw_type'])
                    data = prep_shlexed(data)
                    # print (f"column: {c_name}")
                    if data['name'] in file_data['primary_keys']:
                        data['is_primary_key'] = True
                    data = parse_col_type(data)
                    data = parse_allow_null(data)
                    data = parse_col_comment(data)



                    # del data['shlexed']
                    cols[c_name] = data
                # print ("---"*20)
                break
    # cols = determine_primary_column(cols)
    file_data['columns'] = cols
    # cfu.write.to_json("tmp.sql.delete.json",file_data)
    return file_data

def prep_shlexed(col_data):
    '''
        Prepares the shlexed property for parsing by combining indices as necessary.

        ----------

        Arguments
        -------------------------
        `col_data` {dict}
            The column_data dictionary produced by parse_table_columns
            Must include the shlexed property.

        Return {dict}
        ----------------------
        The col_data dictionary with an updated shlexed property.


        Meta
        ----------
        `author`: Colemen Atwood
        `created`: 05-31-2022 07:55:19
        `memberOf`: TableManager
        `version`: 1.0
        `method_name`: prep_shlexed
        @xxx [05-31-2022 07:56:12]: documentation for prep_shlexed
    '''


    output = []
    previous_val = None
    # print(f"col_data['shlexed']: {col_data['shlexed']}")
    for _,d in enumerate(col_data['shlexed']):
        dl = d.lower()
        if previous_val == "not" and dl == "null":
            output[-1] = "not null"
            continue
        if previous_val == "default":
            output[-1] = f"default {d}"
            continue
        if previous_val == "comment":
            output[-1] = f"comment {d}"
            continue
        if col_data['name'] == 'PRIMARY':
            if previous_val == "key":
                output[-1] = f"key {d}"
                continue

        #     if col_data['schlexed']
        output.append(d)
        previous_val = dl
    col_data['shlexed'] = output
    return col_data

def parse_allow_null(col_data):
    '''
        Determines if the column is allowed to be null.

        ----------

        Arguments
        -------------------------
        `col_data` {dict}
            The column_data dictionary produced by parse_table_columns
            Must include the shlexed property.

        Return {dict}
        ----------------------
        The col_data dictionary with an updated allow_nulls property.

        Meta
        ----------
        `author`: Colemen Atwood
        `created`: 05-31-2022 07:29:36
        `memberOf`: TableManager
        `version`: 1.0
        `method_name`: parse_allow_null
        @xxx [05-31-2022 07:31:12]: documentation for parse_allow_null
    '''

    nn_found = False
    for s in col_data['shlexed']:
        if _csu.array_in_string(["not null","NOT NULL"],s,False):
            nn_found = True
    if nn_found is False:
        col_data['allow_nulls'] = True
    return col_data

def parse_col_comment(col_data):
    '''
        Captures a comment related to the column.

        ----------

        Arguments
        -------------------------
        `col_data` {dict}
            The column_data dictionary produced by parse_table_columns
            Must include the shlexed property.


        Return {dict}
        ----------------------
        The col_data dictionary with an updated comment property.

        Meta
        ----------
        `author`: Colemen Atwood
        `created`: 05-31-2022 07:52:33
        `memberOf`: TableManager
        `version`: 1.0
        `method_name`: parse_col_comment
        @xxx [05-31-2022 07:53:42]: documentation for parse_col_comment
    '''


    for s in col_data['shlexed']:
        # s = s.lower()
        # print(f"s: {s}")
        if s.lower().startswith('comment'):
            # _re.replace(r'^comment\s*')
            cmt = s.replace('comment','')
            col_data['comment'] = _csu.strip(cmt,[" "])
            break

    return col_data

def parse_col_type(col_data):
    '''
        Parses the SQL column type from the column data.

        ----------

        Arguments
        -------------------------
        `col_data` {dict}
            The column_data dictionary produced by parse_table_columns
            Must include the shlexed property.

        Return {dict}
        ----------------------
        The col_data dictionary with an updated data_type property.


        Meta
        ----------
        `author`: Colemen Atwood
        `created`: 05-31-2022 07:54:14
        `memberOf`: TableManager
        `version`: 1.0
        `method_name`: parse_col_type
        @xxx [05-31-2022 07:54:59]: documentation for parse_col_type
    '''


    # _sql_data_types = [
    #     "mediumtext",
    #     "mediumblob",
    #     "varbinary",
    #     "timestamp",
    #     "mediumint",
    #     "tinytext",
    #     "tinyblob",
    #     "smallint",
    #     "longtext",
    #     "longblob",
    #     "datetime",
    #     "varchar",
    #     "tinyint",
    #     "integer",
    #     "decimal",
    #     "boolean",
    #     "double",
    #     "double",
    #     "binary",
    #     "bigint",
    #     "float",
    #     "float",
    #     "year",
    #     "time",
    #     "text",
    #     "enum",
    #     "date",
    #     "char",
    #     "bool",
    #     "blob",
    #     "set",
    #     "int",
    #     "dec",
    #     "bit",
    # ]

    for t in col_data['shlexed']:
        for sql_type in _sql_data_types:
            if sql_type in t.lower():
                col_data['data_type'] = sql_type
                break

        col_data['size'] = parse_column_size(t)
        break
    # lower_raw = col_data['raw_type'].lower()
    # for t in _sql_data_types:
    #     if lower_raw.startswith(t):
    #         # print(f"data type: {t}")
    #         col_data['data_type'] = t
    #         break
    return col_data

def parse_column_size(type_string):
    '''
        Called by parse_col_type
        This will capture the column size if the type supports it.

        ----------

        Arguments
        -------------------------
        `type_string` {str}
            The type of the column, example: "varchar(500)"

        Return {int|None}
        ----------------------
        The integer size of the column upon success, None otherwise.

        Meta
        ----------
        `author`: Colemen Atwood
        `created`: 05-31-2022 07:26:51
        `memberOf`: TableManager
        `version`: 1.0
        `method_name`: parse_column_size
        @xxx [05-31-2022 07:28:25]: documentation for parse_column_size
    '''


    sized_cols = [
        "varbinary",
        "mediumint",
        "timestamp",
        "smallint",
        "datetime",
        "varchar",
        "decimal",
        "integer",
        "tinyint",
        "binary",
        "double",
        "double",
        "bigint",
        "float",
        "float",
        "text",
        "char",
        "time",
        "blob"
        "dec",
        "int",
        "bit",
    ]

    lower_raw = type_string.lower()
    for sc in sized_cols:
        if sc in lower_raw:
            pat = _re.compile(sc+r"\(([0-9]*)\)")
            match = _re.findall(pat,lower_raw)
            if match is not None:
                if len(match) > 0:
                    return int(match[0])
    return None

def strip_keys_constraints(sql):
    output = {
        "sql":[],
        "keys":[],
        "constraints":[],
    }
    lines = sql.split("\n")
    for line in lines:
        if len(line) == 0 or is_line_comment(line):
            continue
        line = _sql.escape_quoted_commas(line)
        # print(f"line: {line}")
        if is_line_key(line):
            output['keys'].append(line)
            continue
        if is_line_constraint(line):
            output['constraints'].append(line)
            continue
        # match = _re.findall(r'^(\w+\s*)?key',line,_re.IGNORECASE)
        # match = _re.findall(r'^((\w+\s*)?key|constraint)',line,_re.IGNORECASE)
        output['sql'].append(line)
    output['sql'] = '\n'.join(output['sql'])

    if len(output['keys']) > 0:
        primary_keys = []
        for k in output['keys']:
            data = parse_primary_keys(k)
            if len(data) > 0:
                primary_keys = data
        output['primary_keys'] = primary_keys

    if len(output['constraints']) > 0:
        foreign_keys = []
        for k in output['constraints']:
            data = parse_constraint(k)
            if data is not None:
                foreign_keys.append(data)
                continue
            data = parse_key(k)
            if data is not None:
                foreign_keys.append(data)
                continue
        output['foreign_keys'] = foreign_keys

    return output

def is_line_key(line:str)->bool:
    '''
        Test if the line contains a key constraint or statement.

        ----------

        Arguments
        -------------------------
        `line` {str}
            The string to parse.

        Return {bool}
        ----------------------
        True upon success, false otherwise.

        Meta
        ----------
        `author`: Colemen Atwood
        `created`: 06-05-2022 20:16:42
        `memberOf`: sql_parse
        `version`: 1.0
        `method_name`: is_line_key
        * @xxx [06-05-2022 20:17:55]: documentation for is_line_key
    '''


    match = _re.match(r'^(\w+\s*)?key',line,_re.IGNORECASE)
    if match is not None:
        return True
    return False

def is_line_comment(line:str)->bool:
    '''
        Determine if the line is commented.

        ----------

        Arguments
        -------------------------
        `line` {str}
            The string to parse.

        Return {bool}
        ----------------------
        True if the line is commented, False otherwise.

        Meta
        ----------
        `author`: Colemen Atwood
        `created`: 06-05-2022 20:13:33
        `memberOf`: sql_parse
        `version`: 1.0
        `method_name`: is_line_comment
        * @xxx [06-05-2022 20:14:45]: documentation for is_line_comment
    '''

    match = _re.match(r'^--',line,_re.IGNORECASE)
    if match is not None:
        return True
    return False

def is_line_constraint(line:str)->bool:
    '''
        Determine if the line provided contains a constraint.
        This is used to quickly test the strings before actually parsing, just to save time.

        ----------

        Arguments
        -------------------------
        `line` {str}
            The string to parse.

        Return {bool}
        ----------------------
        True if it contains a constraint, False otherwise.

        Meta
        ----------
        `author`: Colemen Atwood
        `created`: 06-05-2022 20:11:27
        `memberOf`: sql_parse
        `version`: 1.0
        `method_name`: is_line_constraint
        * @xxx [06-05-2022 20:13:17]: documentation for is_line_constraint
    '''


    match = _re.match(r'^constraint',line,_re.IGNORECASE)
    if match is not None:
        return True
    return False

def parse_primary_keys(sql:str)->list:
    '''
        Parse the primary keys from a SQL string.

        ----------

        Arguments
        -------------------------
        `sql` {str}
            The SQL string to parse.

        Return {list}
        ----------------------
        A list of primary keys.

        Meta
        ----------
        `author`: Colemen Atwood
        `created`: 06-05-2022 20:18:46
        `memberOf`: sql_parse
        `version`: 1.0
        `method_name`: parse_primary_keys
        * @xxx [06-05-2022 20:20:08]: documentation for parse_primary_keys
    '''


    primary_keys = []
    match = _re.findall(r'primary\s*key\s*\(([a-zA-Z0-9_,\`\'\"\s]*)\)',sql,_re.IGNORECASE)
    if match is not None:
        if len(match) > 0:
            keys = _sql.strip_sql_quotes(match[0])
            keys = keys.split(",")
            keys = [_csu.strip(x,[" "]) for x in keys]
            primary_keys = keys
        # print(match)
    return primary_keys

def format_sql_string(sql:str):
    '''
        Format an SQL string to a consistent indentation.

        ----------

        Arguments
        -------------------------
        `sql` {string}
            The sql string to mod.

        Return {string}
        ----------------------
        The formatted SQL

        Meta
        ----------
        `author`: Colemen Atwood
        `created`: 03-21-2022 12:38:32
        `version`: 1.0
        `method_name`: format_sql_string
        @xxx [03-21-2022 12:38:41]: documentation for format_sql_string
    '''
    if _f.exists(sql):
        sql = _f.readr(sql)

    sql = _csu.strip_empty_lines(sql)

    raw = sql
    # raw = _re.sub(r'^[\n]*','',raw)
    statements = _sqlparse.split(raw)
    new_contents = []
    for state in statements:
        state = _re.sub(r'^[\s\n]*','',state)
        new_state = _sqlparse.format(state, reindent=True, keyword_case='upper')
        new_contents.append(new_state)
    if len(new_contents) > 0:
        return "\n".join(new_contents)
    return False

def get_statements(sql:str):
    '''
        Parses the SQL statements from the string.

        ----------

        Arguments
        -------------------------
        `sql` {string}
            The sql string to parse.

        Return {list}
        ----------------------
        A list of SQL statements.

        Meta
        ----------
        `author`: Colemen Atwood
        `created`: 03-21-2022 12:40:52
        `version`: 1.0
        `method_name`: get_statements
        @xxx [03-21-2022 12:40:58]: documentation for get_statements
    '''


    raw = sql
    if isinstance(sql,(str)):
        if _f.exists(sql):
            raw = _f.readr(sql)

    raw = _sql.strip_comments(raw)
    raw = _sql.escape_quoted_chars(raw)
    # raw = format_sql_string(raw)
    statements = [x for x in _sqlparse.parse(raw)]
    return statements

def determine_statement_purpose(statement:str):
    '''
        Determine the purpose of the SQL statement.

        ----------

        Arguments
        -------------------------
        `statement` {str}
            The SQL statement to parse.

        Return {str|None}
        ----------------------
        The statement's purpose or None

        Meta
        ----------
        `author`: Colemen Atwood
        `created`: 06-05-2022 20:08:22
        `memberOf`: sql_parse
        `version`: 1.0
        `method_name`: determine_statement_purpose
        * @xxx [06-05-2022 20:10:56]: documentation for determine_statement_purpose
    '''

    # _log("sql_parse.determine_statement_purpose")
    statement_types = ["CREATE DATABASE","ALTER DATABASE","CREATE SCHEMA","CREATE INDEX","CREATE TABLE","ALTER TABLE","INSERT INTO","DROP INDEX","DROP TABLE","DELETE","UPDATE","SELECT",]
    for s in statement_types:
        if s in statement.upper():
            return s

    return None

def parse(sql):
    _log("sql_parse.parse")
    if _f.exists(sql):
        _log(f"    Reading SQL file: {sql}")
        sql = _f.readr(sql)

    if len(sql)== 0:
        return None

    sql = _csu.strip_empty_lines(sql)
    raw_statements = get_statements(sql)
    # print(f"raw_statements: {raw_statements}")
    data = {
        "schemas":[],
        "tables":[],
        "statements":[],
    }

    for s in raw_statements:
        s = s.value
        # print(f"s: {s}")
        state = {
            "raw":s,
            "purpose":determine_statement_purpose(s),
            "data":None,
        }
        if state['purpose'] is None:
            _log(f"    Failed to determine purpose of statement:\n {s}","error")
            continue

        if state['purpose'] == "CREATE TABLE":
            # _log(f"    Create Table Statement found statement:\n {s}","success")
            state['data'] = parse_table_statement(state['raw'])
            _lu.append(data['tables'],state['data'])
            continue

        if state['purpose'] == "CREATE SCHEMA":
            state['data'] = parse_schema_statement(state['raw'])
            _lu.append(data['schemas'],state['data'])
            continue

        data['statements'].append(state)



    return data













# # pylint: disable=missing-function-docstring
# # pylint: disable=missing-class-docstring
# # pylint: disable=line-too-long
# '''
#     Methods for managing default drawio nodes.

#     Most nodes in drawio are represented by mxCell tags, unless they have meta data, then they are wrapped
#     in an object tag.

#     Connectors are mxcell tags with source and target attributes, we use a connector class just to make 
#     the process a little easier.

#     ----------

#     Meta
#     ----------
#     `author`: Colemen Atwood
#     `created`: 06-04-2022 15:45:00
#     `memberOf`: drawio
#     `name`: mxcell
# '''


# from lxml import etree as _etree
# import utils.string_utils as _csu

# # import json
# # import objectUtils as obj
# from utils.drawio.nodeBase import NodeBase as _NodeBase
# import utils.drawio.diagram_utils as _dia
# # from io import StringIO, BytesIO


# # pylint: disable=missing-function-docstring
# # pylint: disable=missing-class-docstring

# class Mxcell(_NodeBase):
#     def __init__(self,tree,element=None,diagram=None):
#         super().__init__(tree,element,diagram)
#         self.settings = {}
#         self.tree = tree
#         self.element = element
#         self.data = {
#             "node_type":"mxcell",
#             "lxml":"",
#             "xml":"",
#             "coords":{
#                 "x":None,
#                 "y":None,
#                 "tlc":[],
#                 "trc":[],
#                 "brc":[],
#                 "blc":[],
#             },
#             "attributes":{},
#         }
#         self._from_element()
#         # self.set_defaults()



#     def _from_element(self):
#         element = self.element
#         if element is not None:

#             # @Mstep [] parse the mxcell attributes
#             # mxcell = element.xpath('mxCell')
#             self.data['attributes'] = _dia.attrib_to_dict(element.attrib)

#             if 'style' in self.data['attributes']:
#                 self.data['attributes']['style'] = _dia.style_to_dict(self.data['attributes']['style'])

#             # @Mstep [] retrieve and parse the mxGeometry attributes.
#             mxgeometry = element.xpath('mxGeometry')
#             if len(mxgeometry) > 0:
#                 self.data['mxgeometry'] = _dia.attrib_to_dict(mxgeometry[0].attrib)



#             return self.data

#     # def set_coords(self,x=None,y=None,w=None,h=None):
#     #     if self.data['coords']['x'] is None:
#     #         self.coords()

#     #     if x is None:
#     #         x = self.data['mxgeometry']['x']
#     #     if y is None:
#     #         y = self.data['mxgeometry']['y']
#     #     if w is None:
#     #         w = self.data['mxgeometry']['width']
#     #     if h is None:
#     #         h = self.data['mxgeometry']['height']

#     #     self.data['coords']['x'] = x
#     #     self.data['coords']['y'] = y
#     #     self.data['coords']['width'] = w
#     #     self.data['coords']['height'] = h

#     #     self.data['mxgeometry']['x'] = self.data['coords']['x']
#     #     self.data['mxgeometry']['y'] = self.data['coords']['y']
#     #     self.data['mxgeometry']['width'] = self.data['coords']['width']
#     #     self.data['mxgeometry']['height'] = self.data['coords']['height']
#     #     mxCell = self.element.xpath('mxCell')
#     #     mxGeo = mxCell[0].xpath('mxGeometry')
#     #     for k,v in self.data['mxgeometry'].items():
#     #         mxGeo[0].attrib[k] = str(v)

#     # def coords(self):
#     #     x = self.data['mxgeometry']['x']
#     #     y = self.data['mxgeometry']['y']
#     #     width = self.data['mxgeometry']['width']
#     #     height = self.data['mxgeometry']['height']

#     #     self.data['coords']['x'] = x
#     #     self.data['coords']['y'] = y
#     #     self.data['coords']['tlc'] = [x,y]
#     #     self.data['coords']['trc'] = [x + width,y]
#     #     self.data['coords']['brc'] = [x + width,y + height]
#     #     self.data['coords']['blc'] = [x,y + height]

#     #     return self.data['coords']


# def new_mxcell(tree,diagram,id:str=None,parent=None):
#     '''
#         Create a new mxcell node in the diagram.

#         ----------

#         Arguments
#         -------------------------
#         `arg_name` {etree}
#             A reference to the etree xml

#         `diagram` {Diagram}
#             A reference to the Diagram instance this node belongs to.
#         [`id`=None] {str}
#             The node id, if not provided a random one is generated.

#         [`parent`=None] {type}
#             ???


#         Return {Mxcell}
#         ----------------------
#         An instance of the Mxcell Class

#         Meta
#         ----------
#         `author`: Colemen Atwood
#         `created`: 06-05-2022 10:51:06
#         `memberOf`: mcxell
#         `version`: 1.0
#         `method_name`: new_mxcell
#         * @TODO []: documentation for new_mxcell
#     '''


#     mxCell = _etree.SubElement(diagram.dia_root, 'mxCell')
#     id = _csu.gen.rand() if id is None else id
#     mxCell.attrib["id"] = id
#     if parent is not None:
#         mxCell.attrib["parent"] = parent

#     return Mxcell(tree,mxCell,diagram)






from colorama import Fore as _Fore
from colorama import Style as _Style


_CONFIG = {
    "verbose":True,
}

def get(key,default_value=None):
    if key in _CONFIG:
        return _CONFIG[key]
    return default_value


def log(message,style=None):
    if get("verbose",False):
        if style is None:
            print(message)
        if style == "error":
            print(_Fore.RED + message + _Style.RESET_ALL)
        if style == "success":
            print(_Fore.GREEN + message + _Style.RESET_ALL)
        

